@@ -2,7 +2,7 @@ import json
 import multiprocessing
 import os
 import pickle
-from functools import partial
+from functools import partial, wraps
 
 from pydub.utils import mediainfo
 
@@ -12,6 +12,18 @@ import audalign.filehandler as filehandler
 import audalign.fingerprint as fingerprint
 import audalign.recognize as recognize
 import audalign.visrecognize as visrecognize
+import audalign.datalign as datalign
+
+
+def add_rankings(func):
+    @wraps(func)
+    def wrapper_decorator(*args, **kwargs):
+        results = func(*args, **kwargs)
+        if results is not None:
+            results["rankings"] = Audalign.rank_alignment(results)
+        return results
+
+    return wrapper_decorator
 
 
 class Audalign:
@@ -57,6 +69,7 @@ class Audalign:
         Accuracy settings are acheived by manipulations in fingerprinting variables.
 
         Args
+        ----
             arg1 (str): Optional file path to load json or pickle file of already fingerprinted files
             multiprocessing (bool): option to turn off multiprocessing
             num_processors (int): number of processors or threads to use for multiprocessing. Uses all if not given
@@ -84,7 +97,8 @@ class Audalign:
     def set_hash_style(self, hash_style: str) -> None:
         """Sets the hash style. Must be one of ["base", "panako", "panako_mod", "base_three"]
 
-        Args:
+        Args
+        ----
             hash_style (str): Method to use for hashing of fingerprints
         """
         if hash_style not in ["base", "panako", "panako_mod", "base_three"]:
@@ -109,6 +123,7 @@ class Audalign:
         Specific values for accuracy levels were chosen semi-arbitrarily from experimentation to give a few good options.
 
         Args
+        ----
             accuracy (int): which accuracy level: 1-4
         """
         accuracy = int(accuracy)
@@ -156,7 +171,8 @@ class Audalign:
     def set_freq_threshold(self, threshold: int) -> None:
         """Sets minimum frequency threshold for fingerprint
 
-        Args:
+        Args
+        ----
             threshold ([int]): [threshold]
         """
         self.freq_threshold = threshold
@@ -178,7 +194,8 @@ class Audalign:
     def set_multiprocessing(self, true_or_false: bool) -> None:
         """Sets to true for on or false for off
 
-        Args:
+        Args
+        ----
             true_or_false (bool): [true on or false off]
         """
         self.multiprocessing = true_or_false
@@ -190,7 +207,8 @@ class Audalign:
     def set_num_processors(self, num_processors: int) -> None:
         """Set to none to use all processors by default if multiprocessing is true
 
-        Args:
+        Args
+        ----
             num_processors (int): number of processors to use or None for all of them
         """
         self.num_processors = num_processors
@@ -203,7 +221,8 @@ class Audalign:
         """
         Serializes fingerprinted files to json or pickle file
 
-        Args:
+        Args
+        ----
             filename (str): file to load saved fingerprints from
         """
 
@@ -222,6 +241,7 @@ class Audalign:
         Loads/adds saved json or pickle file into current audalign object
 
         Args
+        ----
             filename (str): must be either json or pickle extension
 
         Returns
@@ -267,6 +287,7 @@ class Audalign:
         Fingerprints all files in given directory and all subdirectories
 
         Args
+        ----
             path (str): path to directory to be fingerprinted
             plot (boolean): if true, plots the peaks to be fingerprinted on a spectrogram
             extensions (list[str]): specify which extensions to fingerprint
@@ -304,6 +325,7 @@ class Audalign:
         Fingerprints all files in given directory and all subdirectories
 
         Args
+        ----
             path (str): path to directory to be fingerprinted
             plot (boolean): if true, plots the peaks to be fingerprinted on a spectrogram
             extensions (list[str]): specify which extensions to fingerprint
@@ -390,7 +412,8 @@ class Audalign:
         Fingerprints given file and adds to fingerprinted files
 
         Args
-            file_path (str): path to word to be fingerprinted
+        ----
+            file_path (str): path of file to be fingerprinted
             start_end (tuple(float, float), optional): Silences before and after start and end. (0, -1) Silences last second, (5.4, 0) silences first 5.4 seconds
             set_file_name (str): option to set file name manually rather than use file name in file_path
             plot (boolean): if true, plots the peaks to be fingerprinted on a spectrogram
@@ -425,7 +448,8 @@ class Audalign:
         Fingerprints given file and adds to fingerprinted files
 
         Args
-            file_path (str): path to word to be fingerprinted
+        ----
+            file_path (str): path to file to be fingerprinted
             start_end (tuple(float, float), optional): Silences before and after start and end. (0, -1) Silences last second, (5.4, 0) silences first 5.4 seconds
             set_file_name (str): option to set file name manually rather than use file name in file_path
             plot (boolean): if true, plots the peaks to be fingerprinted on a spectrogram
@@ -446,6 +470,7 @@ class Audalign:
         file_name = set_file_name or file_name
         return [file_name, hashes]
 
+    @add_rankings
     def recognize(
         self,
         file_path: str,
@@ -465,7 +490,8 @@ class Audalign:
         Locality option used to only return match results within certain second range
 
         Args
-            file_path (str): file path of target file to recognize
+        ----
+            file_path (str): file path of target file to recognize.
             filter_matches (int): filters all matches lower than given argument, 1 is recommended
             locality (float): filters matches to only count within locality. In seconds
             locality_filter_prop (int, float,optional): within each offset, filters locality tuples by proportion of highest confidence to tuple confidence
@@ -494,6 +520,7 @@ class Audalign:
             **kwargs,
         )
 
+    @add_rankings
     def correcognize_spectrogram(
         self,
         target_file_path: str,
@@ -511,12 +538,13 @@ class Audalign:
     ):
         """Uses cross correlation on spectrograms to find alignment
 
-        Args:
+        Args
+        ----
             target_file_path (str): File to recognize
             against_file_path (str): File to recognize against
             start_end_target (tuple(float, float), optional): Silences before and after start and end. (0, -1) Silences last second, (5.4, 0) silences first 5.4 seconds
             start_end_against (tuple(float, float), optional): Silences before and after start and end. (0, -1) Silences last second, (5.4, 0) silences first 5.4 seconds
-            filter_matches (float, optional): Filters based on confidence. Ranges between 0 and 1. Defaults to 0.5.
+            filter_matches (float, optional): Filters based on confidence. Ranges between 0 and 1. Defaults to 0.0.
             match_len_filter (int, optional): Limits number of matches returned. Defaults to 30.
             locality (float): filters matches to only count within locality. In seconds
             locality_filter_prop (int, float,optional): within each offset, filters locality tuples by proportion of highest confidence to tuple confidence
@@ -525,7 +553,8 @@ class Audalign:
             plot (bool, optional): Plots. Defaults to False.
             kwargs: additional arguments for scipy.signal.find_peaks.
 
-        Returns:
+        Returns
+        -------
             dict: dictionary of recognition information
         """
         return correcognize.correcognize(
@@ -544,6 +573,7 @@ class Audalign:
             **kwargs,
         )
 
+    @add_rankings
     def correcognize(
         self,
         target_file_path: str,
@@ -564,12 +594,13 @@ class Audalign:
         Faster than visrecognize or recognize and more useful for amplitude
         based alignments
 
-        Args:
+        Args
+        ----
             target_file_path (str): File to recognize
             against_file_path (str): File to recognize against
             start_end_target (tuple(float, float), optional): Silences before and after start and end. (0, -1) Silences last second, (5.4, 0) silences first 5.4 seconds
             start_end_against (tuple(float, float), optional): Silences before and after start and end. (0, -1) Silences last second, (5.4, 0) silences first 5.4 seconds
-            filter_matches (float, optional): Filters based on confidence. Ranges between 0 and 1. Defaults to 0.5.
+            filter_matches (float, optional): Filters based on confidence. Ranges between 0 and 1. Defaults to 0.0.
             match_len_filter (int, optional): Limits number of matches returned. Defaults to 30.
             locality (float): filters matches to only count within locality. In seconds
             locality_filter_prop (int, float,optional): within each offset, filters locality tuples by proportion of highest confidence to tuple confidence
@@ -578,7 +609,8 @@ class Audalign:
             plot (bool, optional): Plots. Defaults to False.
             kwargs: additional arguments for scipy.signal.find_peaks.
 
-        Returns:
+        Returns
+        -------
             dict: dictionary of recognition information
         """
         return correcognize.correcognize(
@@ -596,6 +628,7 @@ class Audalign:
             **kwargs,
         )
 
+    @add_rankings
     def visrecognize(
         self,
         target_file_path: str,
@@ -616,7 +649,8 @@ class Audalign:
         Uses multiprocessing if multiprocessing variable is set to true
         Uses audalign freq_threshold as well
 
-        Args:
+        Args
+        ----
             target_file_path (str): File to recognize.
             against_file_path (str): Recognize against.
             start_end_target (tuple(float, float), optional): Silences before and after start and end. (0, -1) Silences last second, (5.4, 0) silences first 5.4 seconds
@@ -656,6 +690,7 @@ class Audalign:
             plot=plot,
         )
 
+    @add_rankings
     def correcognize_directory(
         self,
         target_file_path: str,
@@ -676,7 +711,8 @@ class Audalign:
         Faster than visrecognize or recognize and more useful for amplitude
         based alignments
 
-        Args:
+        Args
+        ----
             target_file_path (str): File to recognize
             against_directory (str): Directory to recognize against
             start_end (tuple(float, float), optional): Silences before and after start and end. (0, -1) Silences last second, (5.4, 0) silences first 5.4 seconds
@@ -690,7 +726,8 @@ class Audalign:
             _file_audsegs (dict, optional): For use with align.
             kwargs: additional arguments for scipy.signal.find_peaks.
 
-        Returns:
+        Returns
+        -------
             dict: dictionary of recognition information
         """
         return correcognize.correcognize_directory(
@@ -710,6 +747,7 @@ class Audalign:
             **kwargs,
         )
 
+    @add_rankings
     def correcognize_spectrogram_directory(
         self,
         target_file_path: str,
@@ -727,11 +765,12 @@ class Audalign:
     ):
         """Uses cross correlation on spectrogram to find alignment
 
-        Args:
+        Args
+        ----
             target_file_path (str): File to recognize
             against_directory (str): Directory to recognize against
             start_end (tuple(float, float), optional): Silences before and after start and end. (0, -1) Silences last second, (5.4, 0) silences first 5.4 seconds
-            filter_matches (float, optional): Filters based on confidence. Ranges between 0 and 1. Defaults to 0.5.
+            filter_matches (float, optional): Filters based on confidence. Ranges between 0 and 1. Defaults to 0.0.
             match_len_filter (int, optional): Limits number of matches returned. Defaults to 30.
             locality (float): filters matches to only count within locality. In seconds
             locality_filter_prop (int, float,optional): within each offset, filters locality tuples by proportion of highest confidence to tuple confidence
@@ -741,7 +780,8 @@ class Audalign:
             _file_audsegs (dict, optional): For use with align.
             kwargs: additional arguments for scipy.signal.find_peaks.
 
-        Returns:
+        Returns
+        -------
             dict: dictionary of recognition information
         """
         return correcognize.correcognize_directory(
@@ -760,6 +800,7 @@ class Audalign:
             **kwargs,
         )
 
+    @add_rankings
     def visrecognize_directory(
         self,
         target_file_path: str,
@@ -780,7 +821,8 @@ class Audalign:
         Uses multiprocessing if multiprocessing variable is set to true
         Uses audalign freq_threshold as well
 
-        Args:
+        Args
+        ----
             target_file_path (str): File to recognize.
             against_directory (str): Recognize against all files in directory.
             start_end (tuple(float, float), optional): Silences before and after start and end. (0, -1) Silences last second, (5.4, 0) silences first 5.4 seconds
@@ -829,6 +871,7 @@ class Audalign:
         writes given file to the destination file after processing for fingerprinting
 
         Args
+        ----
             file_path (str): file path of audio file
             destination_file (str): file path and name to write file to
             start_end (tuple(float, float), optional): Silences before and after start and end. (0, -1) Silences last second, (5.4, 0) silences first 5.4 seconds
@@ -852,6 +895,7 @@ class Audalign:
         Plots the file_path's peak chart
 
         Args
+        ----
             file_path (str): file to plot
             start_end (tuple(float, float), optional): Silences before and after start and end. (0, -1) Silences last second, (5.4, 0) silences first 5.4 seconds
 
@@ -866,6 +910,7 @@ class Audalign:
         Resets audalign object to brand new state
 
         Args
+        ----
             None
 
         Returns
@@ -876,6 +921,7 @@ class Audalign:
         self.fingerprinted_files = []
         self.total_fingerprints = 0
 
+    @add_rankings
     def target_align(
         self,
         target_file: str,
@@ -902,7 +948,8 @@ class Audalign:
         aligns them, and writes them to destination_path if given. Uses fingerprinting by defualt,
         but uses visual recognition if false
 
-        Args:
+        Args
+        ----
             target_file (str): File to find alignments against
             directory_path (str): Directory to align against
             destination_path (str, optional): Directory to write alignments to
@@ -910,7 +957,7 @@ class Audalign:
             write_extension (str, optional): audio file format to write to. Defaults to None.
             technique (str, optional): options are "fingerprints", "visual", "correlation_spectrogram", "correlation"
             strength_stat (str, optional): confidence for fingerprints, ssim for visual, mse or count also work for visual. Defaults to None.
-            filter_matches (int, float, optional): filter matches level for fingerprinting. Defaults to 1.
+            filter_matches (int, float, optional): filters matches for each technique, look at each recognition method for default.
             locality (float, optional): In seconds for fingerprints, only matches files within given window sizes
             locality_filter_prop (int, float, optional): within each offset, filters locality tuples by proportion of highest confidence to tuple confidence
             volume_threshold (float, optional): volume threshold for visual recognition. Defaults to 216.
@@ -923,7 +970,8 @@ class Audalign:
             max_lags (float, optional): Maximum lags in seconds for correlation.
             kwargs: additional arguments for scipy.signal.find_peaks
 
-        Returns:
+        Returns
+        -------
             dict: dict of file name with shift as value along with match info
         """
 
@@ -951,6 +999,7 @@ class Audalign:
             **kwargs,
         )
 
+    @add_rankings
     def align_files(
         self,
         filename_a,
@@ -979,13 +1028,14 @@ class Audalign:
         try align_files(*filepath_list) for more concision if desired
 
         Args
+        ----
             filename_a (str): String of path for alignment
             filename_b (str): String of path for alignment
             *filenames (strs): strings of paths for alignment
             destination_path (str): String of path to write alignments to
             write_extension (str): if given, writes all alignments with given extension (ex. ".wav" or "wav")
             technique (str): either "fingerprints", "visual", "correlation_spectrogram", or "correlation"
-            filter_matches (float): filters based on confidence.
+            filter_matches (int, float, optional): filters matches for each technique, look at each recognition method for default.
             locality (float): Only recognizes against fingerprints in given width. In seconds
             locality_filter_prop (int, float,optional): within each offset, filters locality tuples by proportion of highest confidence to tuple confidence
             cor_sample_rate (int): Sampling rate for correlation
@@ -1026,6 +1076,7 @@ class Audalign:
             **kwargs,
         )
 
+    @add_rankings
     def align(
         self,
         directory_path: str,
@@ -1050,11 +1101,12 @@ class Audalign:
         Finds matches and relative offsets for all files in directory_path, aligns them, and writes them to destination_path
 
         Args
+        ----
             directory_path (str): String of directory for alignment
             destination_path (str): String of path to write alignments to
             write_extension (str): if given, writes all alignments with given extension (ex. ".wav" or "wav")
             technique (str): either "fingerprints", "visual", "correlation_spectrogram", or "correlation"
-            filter_matches (float): filters based on confidence.
+            filter_matches (int, float, optional): filters matches for each technique, look at each recognition method for default.
             locality (float): Only recognizes against fingerprints in given width. In seconds
             locality_filter_prop (int, float,optional): within each offset, filters locality tuples by proportion of highest confidence to tuple confidence
             cor_sample_rate (int): Sampling rate for correlation
@@ -1094,6 +1146,7 @@ class Audalign:
             **kwargs,
         )
 
+    @add_rankings
     def fine_align(
         self,
         results,
@@ -1119,12 +1172,13 @@ class Audalign:
         Finds matches and relative offsets for all files in directory_path, aligns them, and writes them to destination_path
 
         Args
+        ----
             results (dict): results from previous alignments.
             directory_path (str): String of directory for alignment
             destination_path (str): String of path to write alignments to
             write_extension (str): if given, writes all alignments with given extension (ex. ".wav" or "wav")
             technique (str): either "fingerprints", "visual", "correlation_spectrogram", or "correlation"
-            filter_matches (float): filters based on confidence.
+            filter_matches (int, float, optional): filters matches for each technique, look at each recognition method for default.
             max_lags (float, optional): Maximum lags in seconds for correlation.
             locality (float): Only recognizes against fingerprints in given width. In seconds
             locality_filter_prop (int, float,optional): within each offset, filters locality tuples by proportion of highest confidence to tuple confidence
@@ -1143,8 +1197,16 @@ class Audalign:
         -------
             files_shifts (dict{float}): dict of file name with shift as value. Includes match_info, fine_match_info, and names_and_paths
         """
+
+        if results is None:
+            print("No Results to Fine Align")
+            return None
+
         print("Fine Aligning...")
 
+        if results.get("rankings") is not None:
+            results.pop("rankings")
+
         if strength_stat == "confidence":
             _ = results["match_info"]
             _ = _[list(_.keys())[0]]
@@ -1207,6 +1269,27 @@ class Audalign:
 
         return new_results
 
+    @staticmethod
+    def rank_alignment(
+        alignment: dict,
+    ):
+        """
+        Ranks alignments and recognitions.
+        Included as "rankings" in every recognition and alignment
+        Ranks each recognition on a scale of 1-10
+        Ranks are not proof of a good alignment, just a tool to gauge the different confidences and techniques
+
+        Args
+        ----
+            alignment (dict): result from either recognition or alignment
+
+        Returns
+        -------
+            dict: form similar to alignments or matches
+
+        """
+        return datalign.rank_alignment(alignment=alignment)
+
     @staticmethod
     def _write_shifted_files(
         files_shifts: dict,
@@ -1218,6 +1301,7 @@ class Audalign:
         Writes files to destination_path with specified shift
 
         Args
+        ----
             files_shifts (dict{float}): dict with file path as key and shift as value
             destination_path (str): folder to write file to
             names_and_paths (dict{str}): dict with name as key and path as value
@@ -1232,10 +1316,12 @@ class Audalign:
 
         if file_path is not a valid file or is a directory, returns empty dict
 
-        Args:
+        Args
+        ----
             file_path (str): file path to file
 
-        Returns:
+        Returns
+        -------
             dict: dict of tags and values
         """
         return mediainfo(filepath=file_path)
@@ -1248,6 +1334,7 @@ class Audalign:
         Writes file to destination_path with specified shift in seconds
 
         Args
+        ----
             file_path (str): file path of file to shift
             destination_path (str): where to write file to and file name
             offset_seconds (float): how many seconds to shift, can't be negative
@@ -1264,6 +1351,7 @@ class Audalign:
         Convert audio file to type specified in destination path
 
         Args
+        ----
             file_path (str): file path of file to shift
             destination_path (str): where to write file to and file name
             start_end (tuple(float, float), optional): Silences before and after start and end. (0, -1) Silences last second, (5.4, 0) silences first 5.4 seconds
@@ -1272,12 +1360,99 @@ class Audalign:
             filename=file_path, wrdestination=destination_path, start_end=start_end
         )
 
+    @staticmethod
+    def uniform_level_file(
+        file_path: str,
+        destination: str,
+        write_extension: str = None,
+        mode: str = "normalize",
+        width: float = 5,
+        overlap_ratio: float = 0.5,
+        exclude_min_db: float = -70,
+    ) -> None:
+        """
+        Levels the file using either of two methods: normalize or average.
+        "normalize" levels the file by peak volume while "average" levels by average volume.
+
+        This function pairs well with remove_noise. Leveling brings sound events up in volume
+        so they influence alignments. Removing noise cancels noise that might have been boosted
+        by the leveling.
+
+        Args
+        ----
+            file_path (str): path of file to level.
+            destination (str): path of destination to write to. directory or file name.
+            write_extension (str, optional): extention/format for writing.
+            mode (str): either "normalize" or "average".
+            width (float): width in seconds for each leveling.
+            overlap_ratio (float): between 1 and 0. overlapping windows.
+            exclude_min_db (float): less than 0. Doesn't level window with max dBFS lower than this.
+
+        Returns
+        -------
+            None
+        """
+        filehandler._uniform_level(
+            file_path=file_path,
+            destination_name=destination,
+            write_extension=write_extension,
+            mode=mode,
+            width=width,
+            overlap_ratio=overlap_ratio,
+            exclude_min_db=exclude_min_db,
+        )
+
+    def uniform_level_directory(
+        self,
+        directory: str,
+        destination: str,
+        write_extension: str = None,
+        mode: str = "normalize",
+        width: float = 5,
+        overlap_ratio: float = 0.5,
+        exclude_min_db: float = -70,
+    ) -> None:
+        """
+        Levels the file using either of two methods: normalize or average.
+        "normalize" levels the file by peak volume while "average" levels by average volume.
+
+        This function pairs well with remove_noise. Leveling brings sound events up in volume
+        so they influence alignments. Removing noise cancels noise that might have been boosted
+        by the leveling.
+
+        Args
+        ----
+            file_path (str): path of file to level.
+            destination (str): path of destination to write to. directory or file name.
+            write_extension (str, optional): extention/format for writing.
+            mode (str): either "normalize" or "average".
+            width (float): width in seconds for each leveling.
+            overlap_ratio (float): between 1 and 0. overlapping windows.
+            exclude_min_db (float): less than 0. Doesn't level window with max dBFS lower than this.
+
+        Returns
+        -------
+            None
+        """
+        filehandler.uniform_level_directory(
+            directory=directory,
+            destination=destination,
+            write_extension=write_extension,
+            mode=mode,
+            width=width,
+            overlap_ratio=overlap_ratio,
+            exclude_min_db=exclude_min_db,
+            use_multiprocessing=self.multiprocessing,
+            num_processes=self.num_processors,
+        )
+
     @staticmethod
     def remove_noise_file(
         filepath: str,
         noise_start: float,
         noise_end: float,
         destination: str,
+        write_extension: str = None,
         alt_noise_filepath: str = None,
         prop_decrease: float = 1,
         use_tensorflow: bool = False,
@@ -1286,11 +1461,13 @@ class Audalign:
     ):
         """Remove noise from audio file by specifying start and end seconds of representative sound sections. Writes file to destination
 
-        Args:
+        Args
+        ----
             filepath (str): filepath to read audio file
             noise_start (float): positition in seconds of start of noise section
             noise_end (float): position in seconds of end of noise section
-            destination (str): filepath of destination to write to
+            destination (str): filepath of destination to write to, full path or directory
+            write_extension (str): if given, writes all alignments with given extension (ex. ".wav" or "wav")
             alt_noise_filepath (str): path of different file for noise sample
             prop_decrease (float): between 0 and 1. Proportion to decrease noise
             use_tensorflow (bool, optional): Uses tensorflow to increase speed if available. Defaults to False.
@@ -1302,6 +1479,7 @@ class Audalign:
             noise_start,
             noise_end,
             destination,
+            write_extension=write_extension,
             alt_noise_filepath=alt_noise_filepath,
             prop_decrease=prop_decrease,
             use_tensorflow=use_tensorflow,
@@ -1316,6 +1494,7 @@ class Audalign:
         noise_start: float,
         noise_end: float,
         destination_directory: str,
+        write_extension: str = None,
         prop_decrease: float = 1,
         use_tensorflow: bool = False,
         verbose: bool = False,
@@ -1324,12 +1503,14 @@ class Audalign:
         """Remove noise from audio files in directory by specifying start and end seconds of representative sound sections. Writes file to destination directory
         Uses multiprocessing if self.multiprocessing is true
 
-        Args:
+        Args
+        ----
             directory (str): filepath for directory to quiet
             nosie_filepath (str): filepath to read noise file
             noise_start (float): positition in seconds of start of noise section
             noise_end (float): position in seconds of end of noise section
             destination_directory (str): filepath of destination directory to write to
+            write_extension (str): if given, writes all alignments with given extension (ex. ".wav" or "wav")
             prop_decrease (float): between 0 and 1. Proportion to decrease noise
             use_tensorflow (bool, optional): Uses tensorflow to increase speed if available. Defaults to False.
             verbose (bool, optional): Shows several plots of noise removal process. Defaults to False.
@@ -1341,6 +1522,7 @@ class Audalign:
             noise_start,
             noise_end,
             destination_directory,
+            write_extension=write_extension,
             prop_decrease=prop_decrease,
             use_tensorflow=use_tensorflow,
             verbose=verbose,
