@@ -10,7 +10,8 @@ from audalign.fingerprint import DEFAULT_FS
 import noisereduce
 from functools import partial
 
-cant_write_ext = [".mov", ".mp4"]
+cant_write_ext = [".mov", ".mp4", ".m4a"]
+cant_read_ext = [".txt", ".md", ".pkf", ".py", ".pyc"]
 
 
 def find_files(path, extensions=["*"]):
@@ -33,12 +34,20 @@ def find_files(path, extensions=["*"]):
                 yield (p, extension)
 
 
-def create_audiosegment(filepath: str, start_end: tuple = None, sample_rate=DEFAULT_FS):
+def create_audiosegment(
+    filepath: str, start_end: tuple = None, sample_rate=DEFAULT_FS, length=None
+) -> AudioSegment:
     if sample_rate is None:
         sample_rate = DEFAULT_FS
     if os.path.splitext(filepath)[1] in [".txt", ".json"]:
         raise CouldntDecodeError
-    audiofile = AudioSegment.from_file(filepath)
+    if len(filepath) > 0:
+        audiofile = AudioSegment.from_file(filepath)
+    else:
+        if length is None:
+            audiofile = AudioSegment.silent(duration=0, frame_rate=sample_rate)
+        else:
+            audiofile = AudioSegment.silent(duration=length, frame_rate=sample_rate)
     audiofile = audiofile.set_frame_rate(sample_rate)
     audiofile = audiofile.set_sample_width(2)
     audiofile = audiofile.set_channels(1)
@@ -121,6 +130,8 @@ def read(
         frame_rate (int): returns the bit depth
     """
 
+    if os.path.splitext(filename)[1] in cant_read_ext:
+        raise CouldntDecodeError
     audiofile = create_audiosegment(
         filename, start_end=start_end, sample_rate=sample_rate
     )
@@ -132,27 +143,16 @@ def read(
 
 
 def _floatify_data(audio_segment: AudioSegment):
-    data = np.frombuffer(audio_segment._data, np.int16)
-    new_data = np.zeros(len(data))
-    for i in range(len(data)):
-        if data[i] < 0:
-            new_data[i] = float(data[i]) / 32768
-        elif data[i] == 0:
-            new_data[i] = 0.0
-        if data[i] > 0:
-            new_data[i] = float(data[i]) / 32767
-    return new_data
+    data = np.frombuffer(audio_segment._data, np.int16).astype(np.float32)
+    data[np.where(data < 0)] /= 32768
+    data[np.where(data > 0)] /= 32767
+    return data
 
 
 def _int16ify_data(data: array):
-    for i in range(len(data)):
-        if data[i] < 0:
-            data[i] = int(data[i] * 32768)
-        elif data[i] == 0:
-            data[i] = int(0)
-        else:
-            data[i] = int(data[i] * 32767)
-    return data
+    data[np.where(data < 0)] *= 32768
+    data[np.where(data > 0)] *= 32767
+    return data.astype(np.int16)
 
 
 def noise_remove(
@@ -160,6 +160,7 @@ def noise_remove(
     noise_start,
     noise_end,
     destination,
+    write_extension: str = None,
     alt_noise_filepath=None,
     prop_decrease=1,
     use_tensorflow=False,
@@ -179,6 +180,7 @@ def noise_remove(
             (noise_start * DEFAULT_FS) : (noise_end * DEFAULT_FS)
         ]
 
+    print(f"Reducing noise: {filepath}")
     reduced_noise_data = noisereduce.reduce_noise(
         new_data,
         noisy_part,
@@ -188,10 +190,23 @@ def noise_remove(
         **kwargs,
     )
 
-    reduced_noise_data = _int16ify_data(reduced_noise_data)
-    audiofile._data = reduced_noise_data.astype(np.int16)
-    with open(destination, "wb") as file_place:
-        audiofile.export(file_place, format=os.path.splitext(destination)[1][1:])
+    audiofile._data = _int16ify_data(reduced_noise_data)
+    # if you pass in a folder for destination
+    if len(os.path.splitext(destination)[1]) == 0:
+        destination = os.path.join(destination, os.path.basename(filepath))
+    if write_extension is not None:
+        if write_extension[0] != ".":
+            write_extension = "." + write_extension
+        destination_name = os.path.splitext(destination)[0] + write_extension
+        print(f"Writing {destination_name}")
+        with open(destination_name, "wb") as file_place:
+            audiofile.export(
+                file_place, format=os.path.splitext(destination_name)[1][1:]
+            )
+    else:
+        print(f"Writing {destination}")
+        with open(destination, "wb") as file_place:
+            audiofile.export(file_place, format=os.path.splitext(destination)[1][1:])
 
 
 def noise_remove_directory(
@@ -200,6 +215,7 @@ def noise_remove_directory(
     noise_start,
     noise_end,
     destination_directory,
+    write_extension: str = None,
     prop_decrease=1,
     use_tensorflow=False,
     verbose=False,
@@ -219,6 +235,7 @@ def noise_remove_directory(
         noise_section=noise_data,
         destination_directory=destination_directory,
         prop_decrease=prop_decrease,
+        write_extension=write_extension,
         use_tensorflow=use_tensorflow,
         verbose=verbose,
         **kwargs,
@@ -247,6 +264,7 @@ def noise_remove_directory(
 def _remove_noise(
     file_path,
     noise_section=[],
+    write_extension: str = None,
     destination_directory="",
     prop_decrease=1,
     use_tensorflow=False,
@@ -268,25 +286,211 @@ def _remove_noise(
             **kwargs,
         )
 
-        reduced_noise_data = _int16ify_data(reduced_noise_data)
-        audiofile._data = reduced_noise_data.astype(np.int16)
+        audiofile._data = _int16ify_data(reduced_noise_data)
 
         file_name = os.path.basename(file_path)
         destination_name = os.path.join(destination_directory, file_name)
         if os.path.splitext(destination_name)[1].lower() in cant_write_ext:
             destination_name = os.path.splitext(destination_name)[0] + ".wav"
 
-        print(f'Noise reduced for "{file_path}" writing to "{destination_name}"')
-
-        with open(destination_name, "wb") as file_place:
-            audiofile.export(
-                file_place, format=os.path.splitext(destination_name)[1][1:]
-            )
+        if write_extension is not None:
+            if write_extension[0] != ".":
+                write_extension = "." + write_extension
+            destination_name = os.path.splitext(destination_name)[0] + write_extension
+            print(f'Noise reduced for "{file_path}" writing to "{destination_name}"')
+            with open(destination_name, "wb") as file_place:
+                audiofile.export(
+                    file_place, format=os.path.splitext(destination_name)[1][1:]
+                )
+        else:
+            print(f'Noise reduced for "{file_path}" writing to "{destination_name}"')
+            with open(destination_name, "wb") as file_place:
+                audiofile.export(
+                    file_place, format=os.path.splitext(destination_name)[1][1:]
+                )
 
     except CouldntDecodeError:
         print(f"    Coudn't Decode {file_path}")
 
 
+def calc_array_indexes(array_length, width, overlap_ratio):
+    index_list = []
+    if width > array_length:
+        index_list += [0]
+    else:
+        [
+            index_list.append(i)
+            for i in range(
+                0, array_length - int(width), int(width * (1 - overlap_ratio))
+            )
+        ]
+        if (
+            array_length - int(width) not in index_list
+            and array_length - int(width) > 0
+        ):
+            index_list.append(array_length - int(width))
+    return index_list
+
+
+def calc_overlap_array(length, index_list, width):
+    overlap_array = np.zeros(length, dtype=np.float32)
+    for index in index_list:
+        overlap_array[index : index + width] += 1
+    return overlap_array
+
+
+def uniform_level_directory(
+    directory: str,
+    destination: str,
+    write_extension: str = None,
+    mode: str = "normalize",
+    width: float = 5,
+    overlap_ratio=0.5,
+    exclude_min_db=-70,
+    use_multiprocessing=False,
+    num_processes=None,
+):
+    _uniform_level_ = partial(
+        _uniform_level,
+        destination_name=destination,
+        write_extension=write_extension,
+        mode=mode,
+        width=width,
+        overlap_ratio=overlap_ratio,
+        exclude_min_db=exclude_min_db,
+    )
+
+    if use_multiprocessing == True:
+
+        try:
+            nprocesses = num_processes or multiprocessing.cpu_count()
+        except NotImplementedError:
+            nprocesses = 1
+        else:
+            nprocesses = 1 if nprocesses <= 0 else nprocesses
+
+        with multiprocessing.Pool(nprocesses) as pool:
+
+            pool.map(_uniform_level_, (x[0] for x in find_files(directory)))
+
+            pool.close()
+            pool.join()
+    else:
+        for i in (x[0] for x in find_files(directory)):
+            _uniform_level_(i)
+
+
+def _uniform_level(
+    file_path: str,
+    destination_name: str,
+    write_extension: str = None,
+    mode: str = "normalize",
+    width: float = 5,
+    overlap_ratio=0.5,
+    exclude_min_db=-70,
+):
+    assert overlap_ratio < 1 and overlap_ratio >= 0
+    try:
+        print(f"Uniform Leveling: {file_path}")
+        audiofile = create_audiosegment(file_path)
+        audiofile_data = np.frombuffer(audiofile._data, np.int16)
+        width *= DEFAULT_FS
+        if width > len(audiofile_data):
+            width = len(audiofile_data)
+        index_list = calc_array_indexes(
+            len(audiofile_data),
+            width,
+            overlap_ratio,
+        )
+        overlap_array = calc_overlap_array(len(audiofile_data), index_list, width)
+        if mode == "normalize":
+            audiofile._data = level_by_normalize(
+                audiofile_data,
+                index_list,
+                overlap_array,
+                width,
+                exclude_min_db,
+            ).astype(np.int16)
+        elif mode == "average":
+            audiofile._data = level_by_ave(
+                audiofile_data,
+                index_list,
+                overlap_array,
+                width,
+                exclude_min_db,
+            ).astype(np.int16)
+        else:
+            raise ValueError(
+                f'Mode must be either "normalize" or "average", not {mode}'
+            )
+        audiofile = audiofile.normalize()
+
+        file_name = os.path.basename(file_path)
+        if len(os.path.splitext(destination_name)[1]) == 0:
+            destination_name = os.path.join(destination_name, file_name)
+        if os.path.splitext(destination_name)[1].lower() in cant_write_ext:
+            destination_name = os.path.splitext(destination_name)[0] + ".wav"
+
+        if write_extension is not None:
+            if write_extension[0] != ".":
+                write_extension = "." + write_extension
+            destination_name = os.path.splitext(destination_name)[0] + write_extension
+            print(f'Uniform leveled "{file_path}" writing to "{destination_name}"')
+            with open(destination_name, "wb") as file_place:
+                audiofile.export(
+                    file_place, format=os.path.splitext(destination_name)[1][1:]
+                )
+        else:
+            print(f'Uniform leveled "{file_path}" writing to "{destination_name}"')
+            with open(destination_name, "wb") as file_place:
+                audiofile.export(
+                    file_place, format=os.path.splitext(destination_name)[1][1:]
+                )
+    except CouldntDecodeError:
+        print(f"    Coudn't Decode {file_path}")
+
+
+def level_by_normalize(
+    audiofile_data, index_list, overlap_array, width, exclude_min_db
+):
+    new_audio_data = np.zeros(len(audiofile_data), dtype=np.float32)
+    silent_audiosegment = create_audiosegment("", length=width / DEFAULT_FS * 1000)
+    for index in index_list:
+        silent_audiosegment._data = audiofile_data[index : index + width]
+        if silent_audiosegment.max_dBFS < exclude_min_db:
+            continue
+        silent_audiosegment = silent_audiosegment.normalize()
+        new_audio_data[index : index + width] += np.frombuffer(
+            silent_audiosegment._data, dtype=np.int16
+        )
+    new_audio_data /= overlap_array
+    return new_audio_data
+
+
+def level_by_ave(audiofile_data, index_list, overlap_array, width, exclude_min_db):
+    new_audio_data = np.zeros(len(audiofile_data), dtype=np.float32)
+    audiosegment_slicer = create_audiosegment("", length=width / DEFAULT_FS * 1000)
+    average_level_list = []
+
+    for index in index_list:
+        audiosegment_slicer._data = audiofile_data[index : index + width]
+        if audiosegment_slicer.max_dBFS < exclude_min_db:
+            continue
+        average_level_list.append(
+            (index, audiosegment_slicer.dBFS, audiosegment_slicer.max_dBFS)
+        )
+    target_ave = max(average_level_list, key=lambda x: x[2] - x[1])
+    target_ave = target_ave[2] - target_ave[1]
+    for index, ave_db, _ in average_level_list:
+        audiosegment_slicer._data = audiofile_data[index : index + width]
+        audiosegment_slicer = audiosegment_slicer + (-ave_db - target_ave)
+        new_audio_data[index : index + width] += np.frombuffer(
+            audiosegment_slicer._data, dtype=np.int16
+        )
+    new_audio_data /= overlap_array
+    return new_audio_data
+
+
 def shift_get_files(results: dict, sample_rate: int = None):
     names_and_paths = results.pop("names_and_paths")
     temp_a = results.pop("match_info")
